#include "circuit.h"

int Circuit::check_circuit() {

    int sum_value_unsatisfied_cost = 0;
    int sum_fault_need_update_cost = 0;

    int sum_fault_propagated_weight = 0;
    int sum_fault_detected_weight = 0;

    for(Gate* g : gates) {

        // assert(g->fault_detected[0] == g->cal_fault_detected(0));
        // assert(g->fault_detected[1] == g->cal_fault_detected(1));
        // assert(g->fault_propagated_len[0] == g->fault_propagated_len[0]);
        // assert(g->fault_propagated_len[1] == g->fault_propagated_len[1]);

        assert(g->value_satisfied == ( g->cal_value() == g->value ));

        bool changed = 0;
        if(!changed && g->cal_fault_detected(0) != g->fault_detected[0]) changed = 1;
        if(!changed && g->cal_fault_detected(1) != g->fault_detected[1]) changed = 1;
        if(!changed && g->cal_propagate_len(0) != g->fault_propagated_len[0]) changed = 1;
        if(!changed && g->cal_propagate_len(1) != g->fault_propagated_len[1]) changed = 1;

        assert(changed == g->fault_need_update);

        if(!g->value_satisfied) sum_value_unsatisfied_cost += g->value_unsatisfied_cost;
        if(g->fault_need_update) sum_fault_need_update_cost += g->fault_need_update_cost;

        if(g->fault_detected[0]) sum_fault_detected_weight += g->fault_detected_weight[0];
        if(g->fault_detected[1]) sum_fault_detected_weight += g->fault_detected_weight[1];

        sum_fault_propagated_weight += g->fault_propagated_len[0] * g->fault_propagated_weight[0];
        sum_fault_propagated_weight += g->fault_propagated_len[1] * g->fault_propagated_weight[1];
        
    }

    int circuit_score = sum_fault_propagated_weight + sum_fault_detected_weight
         - sum_value_unsatisfied_cost - sum_fault_need_update_cost;

    printf("svuc: %d, sfpw: %d, sfpuc: %d, sfdw: %d, sfduc: %d\n", 
        sum_value_unsatisfied_cost,
        sum_fault_propagated_weight, sum_fault_detected_weight);

    return circuit_score;
}